import turtle as _turtle
from math import sin as _sin
from math import cos as _cos
from math import pi as _pi

from .vector import Vector as _Vector
from .curve import Curve as _Curve
from .interval import Interval as _Interval


class Plotter(_turtle.Turtle):
    """Graph plotter for drawing vectors and curves. Inherits from `turtle.Turtle` \
    class.
        
    Parameters
    ----------
    window_title : str
        Title to display on window. Defaults to "curvipy".
    plotting_speed : int
        Curve plotting speed. Defaults to 11.
    background_color : str
        Background color. Can either be a name (e.g. "Blue") or a hex color code. \
        Defaults to "#F1FAEE".
    curve_color : str
        Curve color. Can either be a name (e.g. "Red") or a hex color code. Defaults \
        to "#457B9D".
    curve_width : int
        Curve width. Defaults to 4.
    vector_color : str
        Vector color. Can either be a name (e.g. "Yellow") or a hex color code. Defaults \
        to "#E63946".
    vector_width : int
        Vector width. Defaults to 3.
    vector_head_size : int
        Size of vectors' head. Defaults to 10.
    show_axis : bool
        If true, x-axis and y-axis are shown. Defaults to True.
    axis_color : str
        Axis color. Can either be a name (e.g. "Green") or a hex color code. Defaults \
        to "#A8DADC".
    axis_width : int
        Axis width. Defaults to 2.
    x_axis_scale : int
        X-axis scale. Defaults to 10.
    y_axis_scale : int
        Y-axis scale. Defaults to 10.
    """

    def __init__(
        self,
        window_title: str = "Curvipy",
        plotting_speed: int = 11,
        background_color: str = "#F1FAEE",
        curve_color: str = "#457B9D",
        curve_width: int = 4,
        vector_color: str = "#E63946",
        vector_width: int = 3,
        vector_head_size: int = 10,
        show_axis: bool = True,
        axis_color: str = "#A8DADC",
        axis_width: int = 2,
        x_axis_scale: int = 10,
        y_axis_scale: int = 10,
    ) -> None:

        _turtle.Turtle.__init__(self)
        self.shapesize(0.1, 0.1, 0.1)
        self.shape("square")
        _turtle.title(window_title)
        _turtle.bgcolor(background_color)

        # Screen attributes
        self.plotting_speed = plotting_speed
        self.background_color = background_color

        # Curves attributes
        self.curve_color = curve_color
        self.curve_width = curve_width

        # Vectors attributes
        self.vector_color = vector_color
        self.vector_width = vector_width
        self.vector_head_size = vector_head_size

        # Axis attributes
        self.axis_color = axis_color
        self.show_axis = show_axis
        self.axis_width = axis_width
        self.x_axis_scale = x_axis_scale
        self.y_axis_scale = y_axis_scale

        if self.show_axis:
            self._draw_axis()

    def _goto_without_drawing(self, position: tuple[int, int]) -> None:
        self.speed("fastest")
        self.up()
        self.goto(position)
        self.down()

    def _goto_drawing(self, position: tuple[int, int]) -> None:
        self.speed(self.plotting_speed)
        self.down()
        self.goto(position)

    def _draw_axis(self) -> None:
        self.width(self.axis_width)
        self.color(self.axis_color)

        w, h = _turtle.screensize()

        # y axis
        self._goto_without_drawing((0, -h * 2))
        self.goto(0, h * 2)

        # x axis
        self._goto_without_drawing((-w * 2, 0))
        self.goto((w * 2, 0))

    def plot_vector(self, vector: _Vector) -> None:
        """Plots the given two-dimensional vector.

        Parameters
        ----------
        vector : Vector
            Vector to be plotted.
        """
        # Check if vector is the cero vector (v = [0, 0])
        if not vector.norm:
            return

        # Turtle settings
        self.color(self.vector_color)
        self.width(self.vector_width)

        # Draw vector
        scaled_tail = (
            vector.tail[0] * self.x_axis_scale,
            vector.tail[1] * self.y_axis_scale,
        )
        self._goto_without_drawing(scaled_tail)
        scaled_head = (
            vector.head[0] * self.x_axis_scale,
            vector.head[1] * self.y_axis_scale,
        )
        self._goto_drawing(scaled_head)

        # Draw vector head
        left_head_vector = (
            self.vector_head_size * _cos(vector.angle + _pi * 5 / 4),
            self.vector_head_size * _sin(vector.angle + _pi * 5 / 4),
        )
        left_head_endpoint = (
            scaled_head[0] + left_head_vector[0],
            scaled_head[1] + left_head_vector[1],
        )

        self._goto_drawing(left_head_endpoint)

        right_head_vector = (
            self.vector_head_size * _cos(vector.angle - _pi * 5 / 4),
            self.vector_head_size * _sin(vector.angle - _pi * 5 / 4),
        )
        right_head_endpoint = (
            scaled_head[0] + right_head_vector[0],
            scaled_head[1] + right_head_vector[1],
        )

        self._goto_without_drawing(scaled_head)
        self._goto_drawing(right_head_endpoint)

    def plot_curve(self, curve: _Curve, interval: _Interval) -> None:
        """Plots the given two-dimensional curve in the specified interval.

        Parameters
        ----------
        curve : Curve
            Curve to be plotted.
        interval : Interval
            Interval from which the curve will be plotted.
        """
        # Turtle settings
        self.color(self.curve_color)
        self.width(self.curve_width)

        # Curve points
        points = curve.points(interval)
        start_point = points[0]
        scaled_start_point = (
            self.x_axis_scale * start_point[0],
            self.y_axis_scale * start_point[1],
        )

        # Draw curve
        self._goto_without_drawing(scaled_start_point)
        for point in points[1:]:
            scaled_point = (
                self.x_axis_scale * point[0],
                self.y_axis_scale * point[1],
            )
            self._goto_drawing(scaled_point)

    def plot_animated_curve(
        self, curve: _Curve, interval: _Interval, samples_per_vector: int
    ) -> None:
        """Plots the given curve by drawing a set of vectors pointing at the curve \
        points and then joining the vector heads.

        Parameters
        ----------
        curve : Curve
            Curve to be plotted.
        interval : Interval
            Interval from which the curve will be plotted.
        samples_per_vector : int
            Number of samples per each vector plotted. The less `samples_per_vector` \
            is, the more vectors are drawn. 
        """
        # Plot vectors:
        for i, vector in enumerate(curve.points(interval)):
            if i % samples_per_vector != 0:
                continue

            self.plot_vector(_Vector(vector))

        # Plot curve:
        self.plot_curve(curve, interval)

    def clean(self) -> None:
        """Removes curves and vectors plotted."""
        self.clear()
        _turtle.bgcolor(self.background_color)
        if self.show_axis:
            self._draw_axis()



class Interval:
    """Interval in which a curve will be plotted.

    The interval is splitted into a list of values in which the curve will be evaluated. \
    These values are defined by the number of `samples` specified. The more samples, the \
    more precise the curve plot is.
    
    Parameters
    ----------
    start : int or float
        Real number in which the interval starts.
    end : int or float
        Real number in which the interval ends.
    samples: int
        Number of values within the interval. The more samples, the more precise the \
        curve plot is.
    """

    def __init__(self, start: int | float, end: int | float, samples: int):
        dx = (end - start) / samples
        self.__interval = [start + dx * i for i in range(samples)]

    def __iter__(self):
        return iter(self.__interval)



import math as _math


class Vector:
    """Two-dimensional vector `v = (v1, v2)`.

    Attributes
    ----------
    head : tuple[int or float, int or float]
        Two-dimensional point at which the vector is pointing.
    tail : tuple[int or float, int or float]
        Two-dimensional point at which the vector starts.
    """

    def __init__(
        self,
        head: tuple[int | float, int | float],
        tail: tuple[int | float, int | float] = (0, 0),
    ):
        self.head = head
        self.tail = tail
        self.components = (head[0] - tail[0], head[1] - tail[1])

    @property
    def norm(self) -> float:
        """Norm of the vector."""
        return _math.sqrt(self.components[0] ** 2 + self.components[1] ** 2)

    @property
    def angle(self) -> float:
        """Angle between the vector and the x-axis in radians. This angle \
        is a real number in the interval [-π, π]. If the angle is positive \
        it indicates an anticlockwise rotation from the x-axis, this means \
        that the vector is on the first or the second quadrant. If the angle \
        is negative it indicates a clockwise rotation from the x-axis, this \
        means that the vector is on the third or fourth quadrant.
        """
        angle = _math.acos(self.components[0] / self.norm)
        return angle if self.components[1] >= 0 else -angle

    def __getitem__(self, i: int) -> int | float:
        return self.components[i]

    def __len__(self) -> int:
        return len(self.components)



from distutils.core import setup


setup(
    name="curvipy",
    packages=["curvipy"],
    version="1.0.0",
    license="MIT",
    description="The Python library for making math animations in a few lines of code.",
    author="Dylan Tintenfich",
    author_email="tintenfichdylan05@gmail.com",
    url="https://github.com/dylannalex/curvipy",
    keywords=["curves", "linear-algebra", "mathematics", "animations"],
    install_requires=["PyYAML==6.0", "turtle==0.0.1"],
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
    ],
)



import typing as _typing
from abc import ABC as _ABC
from abc import abstractmethod as _abstractmethod

from .interval import Interval as _Interval


_number = int | float
_vector = tuple[_number, _number]


class Curve(_ABC):
    """Base class for all two-dimensional curves."""

    @_abstractmethod
    def points(self, interval: _Interval) -> list[_number]:
        """Returns the curve point for each value in the given interval.

        Parameters
        ----------
        interval : Interval
            The interval from which the curve will be plotted.

        Returns
        -------
        list[int or float]
            A list of curve points.
        """
        pass


class Function(Curve):
    """Function that given a real number returns another real number `y = f(x)`.

    Parameters
    ----------
    function : Callable[[int or float], int or float]
        Function that given an integer or float returns another integer or float.
    """

    def __init__(self, function: _typing.Callable[[_number], _number]):
        self.function = function

    def points(self, interval: _Interval) -> list[_number]:
        """Returns the function point for each value in the given interval.

        Parameters
        ----------
        interval : Interval
            The interval from which the function will be plotted.

        Returns
        -------
        list[int or float]
            A list of function points.
        """
        return [(x, self.function(x)) for x in interval]


class ParametricFunction(Curve):
    """Function that given a real number returns a 2-dimensional vector `f(t) = <x(t), y(t)>`.

    Parameters
    ----------
    function : Callable[[int or float], tuple[int or float, int or float]]
                Function that given an integer or float returns a tuple containing two \
                integers or floats.
    """

    def __init__(self, parametric_function: _typing.Callable[[_number], _vector]):
        self.parametric_function = parametric_function

    def points(self, interval: _Interval) -> list[_number]:
        """Returns the parametric function point for each value in the given interval.

        Parameters
        ----------
        interval : Interval
            The interval from which the parametric function will be plotted.

        Returns
        -------
        list[int or float]
            A list of parametric function points.
        """
        return [self.parametric_function(t) for t in interval]


class TransformedCurve(Curve):
    """Applies a linear transformation (defined as a 2x2 matrix) to the given curve.

    Parameters
    ----------
    matrix : tuple[\
            tuple[int or float, int or float],\
            tuple[int or float, int or float]\
            ]
        Linear transformation represented as a 2x2 matrix.
        
    curve : Curve
        Curve to be transformed.
    """

    def __init__(
        self,
        matrix: tuple[
            tuple[_number, _number],
            tuple[_number, _number],
        ],
        curve: Curve,
    ):
        self.__curve = curve
        self.__matrix = matrix

    def points(self, interval: _Interval) -> list[_number]:
        """Returns the curve point for each value in the given interval.

        Parameters
        ----------
        interval : Interval
            The interval from which the transformed curve will be plotted.

        Returns
        -------
        list[int or float]
            A list of transformed curve points.
        """
        points = []
        for point in self.__curve.points(interval):
            points.append(
                [
                    point[0] * self.__matrix[0][0] + point[1] * self.__matrix[0][1],
                    point[0] * self.__matrix[1][0] + point[1] * self.__matrix[1][1],
                ]
            )
        return points
